/*
Copyright (C) 1996-1997 Id Software, Inc.

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// save/load routines for several texture formats

#include "textures.h"

byte* LoadPcxTexture(char *filename, int width, int height)
{
	FILE *fin;

	pcx_t	*pcx, pcxbuf;
	byte	palette[768];
	byte	*pix, *pcxrgb;
	int		x, y;
	int		dataByte, runLength;
	int		count;

//
// parse the PCX file
//
	fin = fopen(filename, "rb");

	if (fin == NULL)
	{
		return(NULL);
	}

	fread (&pcxbuf, 1, sizeof(pcxbuf), fin);
	pcx = &pcxbuf;
	if (pcx->manufacturer != 0x0a
		|| pcx->version != 5
		|| pcx->encoding != 1
		|| pcx->bits_per_pixel != 8
		|| pcx->xmax >= 320
		|| pcx->ymax >= 256)
	{
		Con_Printf ("Bad pcx file\n");
		fclose(fin);
		return(NULL);
	}

	// checks dimension only if different from -1
	if ((width != -1) && (height != -1))
	{
		if ((pcx->xmax != width) || (pcx->ymax != height))
		{
			Con_Printf("LoadPcxTexture: Bad dimensions for \"%s\"\n", filename);
			fclose(fin);
			return(NULL);
		}
	}

	// seek to palette
	fseek (fin, -768, SEEK_END);
	fread (palette, 1, 768, fin);

	fseek (fin, sizeof(pcxbuf) - 4, SEEK_SET);

	count = (pcx->xmax+1) * (pcx->ymax+1);
	pcxrgb = malloc( count * 4);

	for (y=0 ; y<=pcx->ymax ; y++)
	{
		pix = pcxrgb + 4*y*(pcx->xmax+1);
		for (x=0 ; x<=pcx->ymax ; )
		{
			dataByte = fgetc(fin);

			if((dataByte & 0xC0) == 0xC0)
			{
				runLength = dataByte & 0x3F;
				dataByte = fgetc(fin);
			}
			else
				runLength = 1;

			while(runLength-- > 0)
			{
				pix[0] = palette[dataByte*3];
				pix[1] = palette[dataByte*3+1];
				pix[2] = palette[dataByte*3+2];
				pix[3] = 255;
				pix += 4;
				x++;
			}
		}
	}

	fclose(fin);
	return(pcxrgb);
}


int freadLittleShort (FILE *f)
{
	byte	b1, b2;

	b1 = fgetc(f);
	b2 = fgetc(f);

	return (short)(b1 + b2*256);
}

int freadLittleLong (FILE *f)
{
	byte	b1, b2, b3, b4;

	b1 = fgetc(f);
	b2 = fgetc(f);
	b3 = fgetc(f);
	b4 = fgetc(f);

	return b1 + (b2<<8) + (b3<<16) + (b4<<24);
}


byte* LoadTgaTexture(char *filename, int width, int height)
{
	FILE	*fin;
	tga_t	header;
	byte	*pixbuf, *targargba;
	int		row, column;
	int		columns, rows, numPixels;

	fin = fopen(filename, "rb");

	if (fin == NULL)
	{
		return(NULL);
	}
	
	header.id_length = fgetc(fin);
	header.colormap_type = fgetc(fin);
	header.image_type = fgetc(fin);	
	header.colormap_index = freadLittleShort(fin);
	header.colormap_length = freadLittleShort(fin);
	header.colormap_size = fgetc(fin);
	header.x_origin = freadLittleShort(fin);
	header.y_origin = freadLittleShort(fin);
	header.width = freadLittleShort(fin);
	header.height = freadLittleShort(fin);
	header.pixel_size = fgetc(fin);
	header.attributes = fgetc(fin);

	if ((header.image_type != 2) && (header.image_type != 10))
	{
		Con_Printf("LoadTgaTexture: Only type 2 and 10 targa RGB images supported\n");
		fclose(fin);
		return(NULL);
	}

	if ((header.pixel_size != 24) && (header.pixel_size != 32))
	{
		Con_Printf("LoadTgaTexture: Only type 24 or 32 bit images supported\n");
		fclose(fin);
		return(NULL);
	}

	if ((width != -1) && (height != -1))
	{
		if ((header.width != width) || (header.height != height))
		{
			Con_Printf("LoadTgaTexture: Invalid dimensions\n");
			fclose(fin);
			return(NULL);
		}
	}

	columns = header.width;
	rows = header.height;
	numPixels = columns * rows;
	targargba = malloc(numPixels * 4);

	// skip TARGA image comment
	if (header.id_length != 0)
	{
		fseek(fin, header.id_length, SEEK_CUR);
	}

	if (header.image_type == 2)
	{
	// Uncompressed, RGB images
		for (row = rows -1; row >= 0; row--)
		{
			pixbuf = targargba + row*columns*4;
			for(column = 0; column<columns; column++) 
			{
				unsigned char red,green,blue,alphabyte;
				switch (header.pixel_size) 
				{
					case 24:							
						blue = getc(fin);
						green = getc(fin);
						red = getc(fin);
						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;
						*pixbuf++ = 255;
					break;
					case 32:
						blue = getc(fin);
						green = getc(fin);
						red = getc(fin);
						alphabyte = getc(fin);
						*pixbuf++ = red;
						*pixbuf++ = green;
						*pixbuf++ = blue;
						*pixbuf++ = alphabyte;
					break;
				}
			}

		}
	}
	else if (header.image_type == 10)
	{
	// Runlength encoded RGB images
		unsigned char red, green, blue, alphabyte, packetHeader, packetSize, j;
		for(row=rows-1; row>=0; row--) 
		{
			pixbuf = targargba + row*columns*4;
			for( column=0; column<columns; ) 
			{
				packetHeader=getc(fin);
				packetSize = 1 + (packetHeader & 0x7f);
				if (packetHeader & 0x80) 
				{        
				// run-length packet
					switch (header.pixel_size) 
					{
						case 24:
							blue = getc(fin);
							green = getc(fin);
							red = getc(fin);
							alphabyte = 255;
						break;
						case 32:
							blue = getc(fin);
							green = getc(fin);
							red = getc(fin);
							alphabyte = getc(fin);
						break;
					}

					for(j=0;j<packetSize;j++) 
					{
						*pixbuf++=red;
						*pixbuf++=green;
						*pixbuf++=blue;
						*pixbuf++=alphabyte;
						column++;
						if (column==columns) 
						{ // run spans across rows
							column=0;
							if (row>0)
								row--;
							else
								goto breakOut;

							pixbuf = targargba + row*columns*4;
						}
					}
				}
				else 
				{                            
				// non run-length packet
					for(j=0;j<packetSize;j++) 
					{
						switch (header.pixel_size) 
						{
							case 24:
								blue = getc(fin);
								green = getc(fin);
								red = getc(fin);
								*pixbuf++ = red;
								*pixbuf++ = green;
								*pixbuf++ = blue;
								*pixbuf++ = 255;
							break;
							case 32:
								blue = getc(fin);
								green = getc(fin);
								red = getc(fin);
								alphabyte = getc(fin);
								*pixbuf++ = red;
								*pixbuf++ = green;
								*pixbuf++ = blue;
								*pixbuf++ = alphabyte;
							break;
						}
						column++;
						if (column==columns) 
						{ 
						// pixel packet run spans across rows
							column=0;
							if (row>0)
								row--;
							else
								goto breakOut;
							pixbuf = targargba + row*columns*4;
						}						
					}
				}
			}

			breakOut:;
		}
	}

	fclose(fin);
	return(targargba);
}

